#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/moduleparam.h>
#include <net/arp.h>

//#include "crypto.h"

#define ERR(...) printk( KERN_ERR "! "__VA_ARGS__ )
#define LOG(...) printk( KERN_INFO "! "__VA_ARGS__ )

#define ifname "virt"

static char* link = "eth0";
module_param( link, charp, 0 );

//static char* ifname = "virt";
//module_param( ifname, charp, 0 );

static struct net_device *crypto_dev = NULL;

struct crypto_priv {
    struct net_device_stats priv_stats;
    struct net_device *slave;
};

rx_handler_result_t handle_frame( struct sk_buff **pskb ) {
   struct sk_buff *skb = *pskb;
   if( crypto_dev ) {
      struct crypto_priv *priv = netdev_priv(crypto_dev);
      priv->priv_stats.rx_packets++;
      priv->priv_stats.rx_bytes += skb->len;
      LOG( "rx: injecting frame from %s to %s", skb->dev->name, crypto_dev->name );
      skb->dev = crypto_dev;
      /* netif_receive_skb(skb);
      return RX_HANDLER_CONSUMED; */
      return RX_HANDLER_ANOTHER;
   }
   return RX_HANDLER_PASS;
}

int crypto_open( struct net_device *dev ) {
   LOG( "%s: device opened", dev->name );
   netif_start_queue( dev );
   return 0;
}

int crypto_close( struct net_device *dev ) {
   LOG( "%s: device closed", dev->name );
   netif_stop_queue( dev );
   return 0;
}

netdev_tx_t crypto_xmit( struct sk_buff *skb, struct net_device *dev ) {
   struct crypto_priv *priv = netdev_priv( dev );
   LOG( "%s: crypto_xmit", dev->name );
   priv->priv_stats.tx_packets++;
   priv->priv_stats.tx_bytes += skb->len;
   if( priv->slave ) {
      skb->dev = priv->slave;
      skb->priority = 1;
      LOG( "tx: injecting frame from %s to %s\n", dev->name, skb->dev->name );
      dev_queue_xmit (skb);
      return 0;
   }
   return NETDEV_TX_OK;
}

struct net_device_stats *crypto_get_stats( struct net_device *dev ) {
   return &((struct crypto_priv *)netdev_priv(dev))->priv_stats;
}

static struct net_device_ops crypto_net_device_ops = {
   .ndo_open = crypto_open,
   .ndo_stop = crypto_close,
   .ndo_get_stats = crypto_get_stats,
   .ndo_start_xmit = crypto_xmit,
};

// #define MAX_ADDR_LEN    32  <netdev.h>
// #define ETH_ALEN        6               /* Octets in one ethernet addr   */ <if_ether.h>

void crypto_setup( struct net_device *dev ) {
   int j;
   ether_setup( dev );
   memset( netdev_priv(dev), 0, sizeof( struct crypto_priv ) );
   dev->netdev_ops = &crypto_net_device_ops;
   // Fill in the MAC address with a phoney 
   for( j = 0; j < ETH_ALEN; ++j )
      dev->dev_addr[ j ] = (char)j;
}

int __init crypto_init( void ) {
   int err = 0;
   struct crypto_priv *crypto_priv;
   crypto_dev = alloc_netdev( sizeof( struct crypto_priv ), ifname"%d", crypto_setup );
   if( crypto_dev == NULL ) {
      ERR( "%s: allocate error", THIS_MODULE->name ); return -ENOMEM;
   }
   crypto_priv = netdev_priv( crypto_dev );
   /* interface name */
   crypto_priv->slave = __dev_get_by_name( &init_net, link );
   if( !crypto_priv->slave ) {
      ERR( "%s: no such net: %s", THIS_MODULE->name, link ); return -ENODEV;
   }
   if( crypto_priv->slave->type != ARPHRD_ETHER && crypto_priv->slave->type != ARPHRD_LOOPBACK ) {
      ERR( "%s: illegal net type", THIS_MODULE->name ); return -EINVAL;
   }
   /* also, and clone its IP, MAC and other information */
   memcpy( crypto_dev->dev_addr, crypto_priv->slave->dev_addr, ETH_ALEN );
   memcpy( crypto_dev->broadcast, crypto_priv->slave->broadcast, ETH_ALEN );
   if( ( err = dev_alloc_name( crypto_dev, crypto_dev->name ) ) ) {
      ERR( "%s: allocate name, error %i\n", THIS_MODULE->name, err );
      return -EIO;
   }
   register_netdev( crypto_dev );
   rtnl_lock();
   netdev_rx_handler_register( crypto_priv->slave, &handle_frame, NULL );
   rtnl_unlock();
   LOG( "register rx handler for %s", crypto_priv->slave->name );
   LOG( "module %s loaded\n", THIS_MODULE->name );
   return 0;
}

void __exit crypto_exit( void ) {
   struct crypto_priv *priv;
   if( crypto_dev ) {
      priv = netdev_priv( crypto_dev );
      if( priv->slave ) {
         LOG( "unregister rx handler for %s\n", priv->slave->name );
         rtnl_lock();
         netdev_rx_handler_unregister( priv->slave );
         rtnl_unlock();
      }
      unregister_netdev( crypto_dev );
      free_netdev( crypto_dev );
   }
   LOG( "module %s unloaded\n", THIS_MODULE->name );
}

module_init( crypto_init );
module_exit( crypto_exit );
MODULE_LICENSE( "GPL" );

